==> xqf_0.8.5.orig.tar.gz: xqf-0.8.5/srv-info.c <==
/* XQF - Quake server browser and launcher
 * Copyright (C) 1998 Roman Pozlevich <roma@botik.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <sys/types.h>
#include <stdio.h>	/* sprintf */
#include <stdlib.h>	/* strtoul, qsort */
#include <string.h>	/* strcmp */

#include <gtk/gtk.h>

#include "xqf.h"
#include "stat.h"
#include "pixmaps.h"
#include "srv-info.h"


#define	QW_TEAMPLAY_CTF 	11

static char *qw_teamplay_ctf[QW_TEAMPLAY_CTF] = {
  "Health Protect",    		/*      1 */
  "Armor Protect",		/*      2 */
  "Damage to Attacker",		/*      4 */
  "Frag Penalty",		/*      8 */
				            
  "Death Penalty",		/*     16 */
  "Team Color Lock",		/*     32 */
  "Static Teams",		/*     64 */
  "Drop Items",			/*    128 */
				            
  "Capture the Flag",		/*    256 */
  "Custom Capture the Flag",	/*    512 */
  "Team Selection"		/*   1024 */
};


#define	QW_TEAMPLAY_TF 		19

static char *qw_teamplay_fortress[QW_TEAMPLAY_TF] = {
  "Teamplay",    				/*      1 */
  "1/2 damage from direct fire",		/*      2 */
  "No damage from direct fire",			/*      4 */
  "1/2 damage from area-affect weaponry",	/*      8 */
				            
  "No damage from area-affect weaponry",	/*     16 */
  "Advantage for team with less members",	/*     32 */
  "Advantage for team with lower score",	/*     64 */
  "1/2 armor damage from direct fire",		/*    128 */

  "No armor damage from direct fire",		/*    256 */
  "1/2 armor damage from area-affect weapon",	/*    512 */
  "No armor damage from area-affect weaponry",	/*   1024 */
  "1/2 mirror damage from direct fire",		/*   2048 */

  "Full mirror damage from direct fire",	/*   4096 */
  "1/2 mirror damage from area-affect weaponry",/*   8192 */
  "Full mirror damage from area-affect weaponry"/*  16384 */
};


#define Q2_DMFLAGS	20

static char *q2_dmflags[Q2_DMFLAGS] = {
  "No Health",                  /*      1 */
  "No Powerups",		/*      2 */
  "Weapons Stay",		/*      4 */
  "No Falling Damage",		/*      8 */
				            
  "Instant Powerups",		/*     16 */
  "Same Map",			/*     32 */
  "Teams by Skin",		/*     64 */
  "Teams by Model",		/*    128 */
				            
  "No Friendly Fire",		/*    256 */
  "Spawn Farthest",		/*    512 */
  "Force Respawn",		/*   1024 */
  "No Armor",			/*   2048 */
				            
  "Allow Exit",			/*   4096 */
  "Infinite Ammo",		/*   8192 */
  "Quad Drop",			/*  16384 */
  "Fixed FOV",		 	/*  32768 */

  NULL,				/*  65536 */
  "CTF Forced Join",	   	/* 131072 */
  "Armor Protect",	   	/* 262144 */
  "CTF No Tech Powerups"  	/* 524288 */
};


#define Q2_GXMOD_DMFLAGS	18

static char *q2_gxmod_dmflags[Q2_GXMOD_DMFLAGS] = {
  "No Health",                  /*      1 */
  "No Powerups",		/*      2 */
  "Weapons Stay",		/*      4 */
  "No Falling Damage",		/*      8 */

  "Instant Powerups",		/*     16 */
  "Same Map",			/*     32 */
  "Teams by Skin",		/*     64 */
  "Teams by Model",		/*    128 */

  "No Friendly Fire",		/*    256 */
  "Spawn Farthest",		/*    512 */
  "Force Respawn",		/*   1024 */
  "No Armor",			/*   2048 */

  "Allow Exit",			/*   4096 */
  "Infinite Ammo",		/*   8192 */
  "Quad Drop",			/*  16384 */
  "Fixed FOV",		 	/*  32768 */

  "Team DM",			/*  65536 */
  "Pratice Mode"	   	/* 131072 */
};


#define Q2_GXMOD_WEAPFLAGS	9

static char *q2_gxmod_weapflags[Q2_GXMOD_WEAPFLAGS] = {
  "Shotgun",     	/*      1 */
  "Super Shotgun",	/*      2 */
  "Machinegun",		/*      4 */
  "Chaingun",		/*      8 */

  "Grenade Launcher",	/*     16 */
  "Rocket Launcher",	/*     32 */
  "Hyperblaster",	/*     64 */
  "Railgun",		/*    128 */

  "BFG"			/*    256 */
};


#define Q2_GXMOD_POWERUPFLAGS	10

static char *q2_gxmod_powerupflags[Q2_GXMOD_POWERUPFLAGS] = {
  "Body Armor",         /*      1 */
  "Combat Armor",	/*      2 */
  "Jacket Armor",	/*      4 */
  "Armor Shard",	/*      8 */

  "Power Shield",	/*     16 */
  "Quad Damage",	/*     32 */
  "Mega Health",	/*     64 */
  "Large Health",	/*    128 */

  "Health",		/*    256 */
  "Small Health"	/*    512 */
};


#define Q2_LMCTF_CTFFLAGS	10

static char *q2_lmctf_ctfflags[Q2_LMCTF_CTFFLAGS] = {
  "Weapon Balancing",		/*      1 */
  "Allow Invulnerability",	/*      2 */
  "Reset Teams Every Level",    /*      4 */
  "No Team Switch",		/*      8 */

  "Offhand Hook",		/*     16 */
  "No Voices",             	/*     32 */
  "No Damage From Grapple",     /*     64 */
  "No Teams (Deathmatch)",     	/*    128 */

  "No Flags (Deathmatch)",	/*    256 */
  "Score Balance"		/*    512 */
};


#define Q2_LMCTF_RUNES		4

static char *q2_lmctf_runes[Q2_LMCTF_RUNES] = {
  "Damage Powerup",		/*      1 */
  "Resist Powerup",	  	/*      2 */
  "Haste Powerup",    		/*      4 */
  "Regen Powerup"	  	/*      8 */
};


#define Q2_EXPERT_EXPFLAGS	17

static char *q2_expert_expflags[Q2_EXPERT_EXPFLAGS] = {
  "Expert Weapons",		/*      1 */
  "Balanced Items",		/*      2 */
  "Free Gear",			/*      4 */
  "Expert Powerups",		/*      8 */

  "No Powerups",		/*      16 */
  "Expert Hook",		/*      32 */
  "No Hacks",			/*      64 */
  "Player ID",			/*      128 */

  "Enforced Teams",		/*      256 */
  "Fair Teams",			/*      512 */
  "No Team Switch",		/*     1024 */
  "Pogo",			/*     2048 */

  NULL,				/*     4096 */
  "Sky Hook",			/*     8192 */
  "No Plats",			/*    16384 */
  "Team Distribution",		/*    32768 */

  "Alternate Restore"		/*    65536 */
};


#define Q2_PBALL_PBFLAGS	8

static char *q2_pball_pbflags[Q2_PBALL_PBFLAGS] = {
  "No Sports Mode",		/*      1 */
  "No Team Change Allowed",	/*      2 */
  "Normal Limits",		/*      4 */
  "No Spectators",		/*      8 */

  "No Map List",		/*      16 */
  "No Orbs",			/*      32 */
  "No HatTrick Mode",		/*      64 */
  "No Assists",			/*      128 */
};


#define Q2_WF_WFFLAGS	14

static char *q2_wf_wfflags[Q2_WF_WFFLAGS] = {
  "Friendly Fire",		/*      1 */
  "Player Classes (unused)",	/*      2 */
  "No Fort Respawn",		/*      4 */
  "No Homing Rockets",		/*      8 */

  "No Flying",			/*      16 */
  "Decoy Pursue",		/*      32 */
  "No Railgun Effect",		/*      64 */
  "No Turrets",			/*      128 */

  "No Earthquake Grenades",	/*      256 */
  NULL,				/*      512 */
  "Next Map Voting",		/*     1024 */
  NULL,				/*     2048 */

  "CTF Mode",			/*     4096 */
  "No Player Classes",		/*     8192 */
};


static void show_extended_flags (const char *str, char *names[], int size,
                                                                int showall) {
  unsigned long flags;
  unsigned long mask;
  char *text[2];
  char buf[32];
  int row;
  int i;

  if (!str || !*str)
    return;

  flags = strtoul (str, NULL, 10);
  if (flags == 0 || flags == ULONG_MAX)
    return;

  mask = 1;

  for (i = 0; flags || (showall && i < size); i++) {
    if ((showall && names[i]) || (flags & mask) != 0) {
      text[0] = NULL;
      text[1] = (i < size && names[i])? names[i] : "?????";

      row = gtk_clist_append (GTK_CLIST (serverinfo_clist), text);
      gtk_clist_set_shift (GTK_CLIST (serverinfo_clist), row, 0, 0, 12);

      sprintf (buf, "%lu", mask);

      gtk_clist_set_pixtext (GTK_CLIST (serverinfo_clist), row, 0, buf, 4, 
                     ((flags & mask) != 0)? gplus_pix : rminus_pix,
                     ((flags & mask) != 0)? gplus_pix_mask : rminus_pix_mask);
    }

    flags &= ~mask;
    mask <<= 1;
  }
}


static void sort_serverinfo (char **info) {
  char **ptr;
  int n = 0;

  if (!info)
    return;

  for (ptr = info; ptr[0] || ptr[1]; ptr += 2)
    n++;

  qsort (info, n, sizeof (char *) * 2, qsort_serverinfo);
}


void populate_serverinfo_clist (struct server *s) {
  char *text[2];
  char **info = s->info;

  if (info == NULL) {
    gtk_clist_clear (GTK_CLIST (serverinfo_clist));
    return;
  }

  gtk_clist_freeze (GTK_CLIST (serverinfo_clist));
  gtk_clist_clear (GTK_CLIST (serverinfo_clist));

  sort_serverinfo (info);

  while (info[0]) {
    text[0] = (info[0])? info[0] : NULL;
    text[1] = (info[1])? info[1] : NULL;

    gtk_clist_append (GTK_CLIST (serverinfo_clist), text);

    switch (s->type) {

    case QW_SERVER:
      if (info[0] && !strcmp (info[0], "teamplay")) {
	if (s->game && (!g_strcasecmp (s->game, "ctf") ||
                                       !g_strcasecmp (s->game, "starwars"))) {
	  show_extended_flags (info[1], qw_teamplay_ctf, QW_TEAMPLAY_CTF, 
                                                                       FALSE);
	}
	else if (s->game && !g_strcasecmp (s->game, "fortress")) {
	  show_extended_flags (info[1], qw_teamplay_fortress, QW_TEAMPLAY_TF, 
                                                                       FALSE);
	}
      }
      break;

    case Q2_SERVER:
      if (info[0] && !strcmp (info[0], "dmflags")) {
	if (s->game && !g_strcasecmp (s->game, "gxmod")) {
	  show_extended_flags (info[1], q2_gxmod_dmflags, Q2_GXMOD_DMFLAGS, 
                                                                       FALSE);
	}
	else {
	  show_extended_flags (info[1], q2_dmflags, Q2_DMFLAGS, FALSE);
	}
      }
      else if (info[0] && !strcmp (info[0], "ctfflags")) {
	if (s->game && !g_strcasecmp (s->game, "lmctf")) {
	  show_extended_flags (info[1], q2_lmctf_ctfflags, Q2_LMCTF_CTFFLAGS, 
                                                                       FALSE);
	}
      }
      else if (info[0] && !strcmp (info[0], "runes")) {
	if (s->game && !g_strcasecmp (s->game, "lmctf"))
	  show_extended_flags (info[1], q2_lmctf_runes, Q2_LMCTF_RUNES, TRUE);
      }
      else if (info[0] && !strcmp (info[0], "expflags")) {
	if (s->game && !g_strcasecmp (s->game, "expert")) {
	  show_extended_flags (info[1], q2_expert_expflags,
                                                   Q2_EXPERT_EXPFLAGS, FALSE);
	}
      }
      else if (info[0] && !strcmp (info[0], "pbflags")) {
	if (s->game && !g_strcasecmp (s->game, "pball")) {
	  show_extended_flags (info[1], q2_pball_pbflags, Q2_PBALL_PBFLAGS, 
                                                                       FALSE);
	}
      }
      else if (info[0] && !strcmp (info[0], "weapflags")) {
	if (s->game && (!g_strcasecmp (s->game, "gxmod") || 
                                             !g_strcasecmp (s->game, "gx"))) {
	  show_extended_flags (info[1], q2_gxmod_weapflags, 
                                                    Q2_GXMOD_WEAPFLAGS, TRUE);
	}
      }
      else if (info[0] && !strcmp (info[0], "powerupflags")) {
	if (s->game && !g_strcasecmp (s->game, "gxmod")) {
	  show_extended_flags (info[1], q2_gxmod_powerupflags, 
                                                 Q2_GXMOD_POWERUPFLAGS, TRUE);
	}
      }
      else if (info[0] && !strcmp (info[0], "wfflags")) {
	if (s->game && !g_strcasecmp (s->game, "wf"))
	  show_extended_flags (info[1], q2_wf_wfflags, Q2_WF_WFFLAGS, FALSE);
      }
      break;

    default:
      break;
    }

    info += 2;
  }

  gtk_clist_thaw (GTK_CLIST (serverinfo_clist));
}

==> xqf-0.8.7.1.tar.gz: xqf-0.8.7.1/srv-info.c <==
/* XQF - Quake server browser and launcher
 * Copyright (C) 1998 Roman Pozlevich <roma@botik.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <sys/types.h>
#include <stdio.h>	/* sprintf */
#include <stdlib.h>	/* strtoul, qsort */
#include <string.h>	/* strcmp */

#include <gtk/gtk.h>

#include "xqf.h"
#include "stat.h"
#include "pixmaps.h"
#include "srv-info.h"


#define	QW_TEAMPLAY_CTF 	11

static char *qw_teamplay_ctf[QW_TEAMPLAY_CTF] = {
  "Health Protect",    		/*      1 */
  "Armor Protect",		/*      2 */
  "Damage to Attacker",		/*      4 */
  "Frag Penalty",		/*      8 */
				            
  "Death Penalty",		/*     16 */
  "Team Color Lock",		/*     32 */
  "Static Teams",		/*     64 */
  "Drop Items",			/*    128 */
				            
  "Capture the Flag",		/*    256 */
  "Custom Capture the Flag",	/*    512 */
  "Team Selection"		/*   1024 */
};


#define	QW_TEAMPLAY_TF 		19

static char *qw_teamplay_fortress[QW_TEAMPLAY_TF] = {
  "Teamplay",    				/*      1 */
  "1/2 damage from direct fire",		/*      2 */
  "No damage from direct fire",			/*      4 */
  "1/2 damage from area-affect weaponry",	/*      8 */
				            
  "No damage from area-affect weaponry",	/*     16 */
  "Advantage for team with less members",	/*     32 */
  "Advantage for team with lower score",	/*     64 */
  "1/2 armor damage from direct fire",		/*    128 */

  "No armor damage from direct fire",		/*    256 */
  "1/2 armor damage from area-affect weapon",	/*    512 */
  "No armor damage from area-affect weaponry",	/*   1024 */
  "1/2 mirror damage from direct fire",		/*   2048 */

  "Full mirror damage from direct fire",	/*   4096 */
  "1/2 mirror damage from area-affect weaponry",/*   8192 */
  "Full mirror damage from area-affect weaponry"/*  16384 */
};


#define Q2_DMFLAGS	20

static char *q2_dmflags[Q2_DMFLAGS] = {
  "No Health",                  /*      1 */
  "No Powerups",		/*      2 */
  "Weapons Stay",		/*      4 */
  "No Falling Damage",		/*      8 */
				            
  "Instant Powerups",		/*     16 */
  "Same Map",			/*     32 */
  "Teams by Skin",		/*     64 */
  "Teams by Model",		/*    128 */
				            
  "No Friendly Fire",		/*    256 */
  "Spawn Farthest",		/*    512 */
  "Force Respawn",		/*   1024 */
  "No Armor",			/*   2048 */
				            
  "Allow Exit",			/*   4096 */
  "Infinite Ammo",		/*   8192 */
  "Quad Drop",			/*  16384 */
  "Fixed FOV",		 	/*  32768 */

  NULL,				/*  65536 */
  "CTF Forced Join",	   	/* 131072 */
  "Armor Protect",	   	/* 262144 */
  "CTF No Tech Powerups"  	/* 524288 */
};


#define Q2_XATRIX_DMFLAGS	17

static char *q2_xatrix_dmflags[Q2_XATRIX_DMFLAGS] = {
  "No Health",                  /*      1 */
  "No Powerups",		/*      2 */
  "Weapons Stay",		/*      4 */
  "No Falling Damage",		/*      8 */
				            
  "Instant Powerups",		/*     16 */
  "Same Map",			/*     32 */
  "Teams by Skin",		/*     64 */
  "Teams by Model",		/*    128 */
				            
  "No Friendly Fire",		/*    256 */
  "Spawn Farthest",		/*    512 */
  "Force Respawn",		/*   1024 */
  "No Armor",			/*   2048 */
				            
  "Allow Exit",			/*   4096 */
  "Infinite Ammo",		/*   8192 */
  "Quad Drop",			/*  16384 */
  "Fixed FOV",		 	/*  32768 */

  "DualFire Drop"		/*  65536 */
};


#define Q2_ROGUE_DMFLAGS	21

static char *q2_rogue_dmflags[Q2_ROGUE_DMFLAGS] = {
  "No Health",                  /*      1 */
  "No Powerups",		/*      2 */
  "Weapons Stay",		/*      4 */
  "No Falling Damage",		/*      8 */
				            
  "Instant Powerups",		/*     16 */
  "Same Map",			/*     32 */
  "Teams by Skin",		/*     64 */
  "Teams by Model",		/*    128 */
				            
  "No Friendly Fire",		/*    256 */
  "Spawn Farthest",		/*    512 */
  "Force Respawn",		/*   1024 */
  "No Armor",			/*   2048 */
				            
  "Allow Exit",			/*   4096 */
  "Infinite Ammo",		/*   8192 */
  "Quad Drop",			/*  16384 */
  "Fixed FOV",		 	/*  32768 */

  NULL,				/*  65536 */
  "No Mines",		   	/* 131072 */
  "No Stack Double",	   	/* 262144 */
  "No Nukes",  			/* 524288 */

  "No Spheres" 			/* 1048576 */
};


#define Q2_GXMOD_DMFLAGS	19

static char *q2_gxmod_dmflags[Q2_GXMOD_DMFLAGS] = {
  "No Health",                  /*      1 */
  "No Powerups",		/*      2 */
  "Weapons Stay",		/*      4 */
  "No Falling Damage",		/*      8 */

  "Instant Powerups",		/*     16 */
  "Same Map",			/*     32 */
  "Teams by Skin",		/*     64 */
  "Teams by Model",		/*    128 */

  "No Friendly Fire",		/*    256 */
  "Spawn Farthest",		/*    512 */
  "Force Respawn",		/*   1024 */
  "No Armor",			/*   2048 */

  "Allow Exit",			/*   4096 */
  "Infinite Ammo",		/*   8192 */
  "Quad Drop",			/*  16384 */
  "Fixed FOV",		 	/*  32768 */

  "Team DM",			/*  65536 */
  "Pratice Mode",	   	/* 131072 */
  "Armor Protect"	   	/* 262144 */
};


#define Q2_GXMOD_WEAPFLAGS	9

static char *q2_gxmod_weapflags[Q2_GXMOD_WEAPFLAGS] = {
  "Shotgun",     	/*      1 */
  "Super Shotgun",	/*      2 */
  "Machinegun",		/*      4 */
  "Chaingun",		/*      8 */

  "Grenade Launcher",	/*     16 */
  "Rocket Launcher",	/*     32 */
  "Hyperblaster",	/*     64 */
  "Railgun",		/*    128 */

  "BFG"			/*    256 */
};


#define Q2_GXMOD_POWERUPFLAGS	11

static char *q2_gxmod_powerupflags[Q2_GXMOD_POWERUPFLAGS] = {
  "Body Armor",         /*      1 */
  "Combat Armor",	/*      2 */
  "Jacket Armor",	/*      4 */
  "Armor Shard",	/*      8 */

  "Power Shield",	/*     16 */
  "Quad Damage",	/*     32 */
  "Mega Health",	/*     64 */
  "Large Health",	/*    128 */

  "Health",		/*    256 */
  "Small Health",	/*    512 */
  "Invincibility"	/*   1024 */
};


#define Q2_LMCTF_CTFFLAGS	10

static char *q2_lmctf_ctfflags[Q2_LMCTF_CTFFLAGS] = {
  "Weapon Balancing",		/*      1 */
  "Allow Invulnerability",	/*      2 */
  "Reset Teams Every Level",    /*      4 */
  "No Team Switch",		/*      8 */

  "Offhand Hook",		/*     16 */
  "No Voices",             	/*     32 */
  "No Damage From Grapple",     /*     64 */
  "No Teams (Deathmatch)",     	/*    128 */

  "No Flags (Deathmatch)",	/*    256 */
  "Score Balance"		/*    512 */
};


#define Q2_LMCTF_RUNES		4

static char *q2_lmctf_runes[Q2_LMCTF_RUNES] = {
  "Damage Powerup",		/*      1 */
  "Resist Powerup",	  	/*      2 */
  "Haste Powerup",    		/*      4 */
  "Regen Powerup"	  	/*      8 */
};


#define Q2_EXPERT_EXPFLAGS	17

static char *q2_expert_expflags[Q2_EXPERT_EXPFLAGS] = {
  "Expert Weapons",		/*      1 */
  "Balanced Items",		/*      2 */
  "Free Gear",			/*      4 */
  "Expert Powerups",		/*      8 */

  "No Powerups",		/*      16 */
  "Expert Hook",		/*      32 */
  "No Hacks",			/*      64 */
  "Player ID",			/*      128 */

  "Enforced Teams",		/*      256 */
  "Fair Teams",			/*      512 */
  "No Team Switch",		/*     1024 */
  "Pogo",			/*     2048 */

  NULL,				/*     4096 */
  "Sky Hook",			/*     8192 */
  "No Plats",			/*    16384 */
  "Team Distribution",		/*    32768 */

  "Alternate Restore"		/*    65536 */
};


#define Q2_PBALL_PBFLAGS	8

static char *q2_pball_pbflags[Q2_PBALL_PBFLAGS] = {
  "No Sports Mode",		/*      1 */
  "No Team Change Allowed",	/*      2 */
  "Normal Limits",		/*      4 */
  "No Spectators",		/*      8 */

  "No Map List",		/*      16 */
  "No Orbs",			/*      32 */
  "No HatTrick Mode",		/*      64 */
  "No Assists"			/*      128 */
};


#define Q2_WF_WFFLAGS	14

static char *q2_wf_wfflags[Q2_WF_WFFLAGS] = {
  "Friendly Fire",		/*      1 */
  "Player Classes (unused)",	/*      2 */
  "No Fort Respawn",		/*      4 */
  "No Homing Rockets",		/*      8 */

  "No Flying",			/*      16 */
  "Decoy Pursue",		/*      32 */
  "No Railgun Effect",		/*      64 */
  "No Turrets",			/*      128 */

  "No Earthquake Grenades",	/*      256 */
  NULL,				/*      512 */
  "Next Map Voting",		/*     1024 */
  NULL,				/*     2048 */

  "CTF Mode",			/*     4096 */
  "No Player Classes"		/*     8192 */
};


#define Q2_BOTS_TEAMFLAGS	7

static char *q2_bots_teamflags[Q2_BOTS_TEAMFLAGS] = {
  "Health Protect",    		/*      1 */
  "Armor Protect",		/*      2 */
  "Damage to Attacker",		/*      4 */
  "Frag Penalty",		/*      8 */
				            
  "Death Penalty",		/*     16 */
  "No Team Change Across Levels",		/*     32 */
  "No Splash Damage from Own Explosives"	/*     64 */
};


static void show_extended_flags (const char *str, char *names[], int size,
                                                                int showall) {
  unsigned long flags;
  unsigned long mask;
  char *text[2];
  char buf[32];
  int row;
  int i;

  if (!str || !*str)
    return;

  flags = strtoul (str, NULL, 10);
  if (flags == 0 || flags == ULONG_MAX)
    return;

  mask = 1;

  for (i = 0; flags || (showall && i < size); i++) {
    if ((showall && names[i]) || (flags & mask) != 0) {
      text[0] = NULL;
      text[1] = (i < size && names[i])? names[i] : "?????";

      row = gtk_clist_append (GTK_CLIST (serverinfo_clist), text);
      gtk_clist_set_shift (GTK_CLIST (serverinfo_clist), row, 0, 0, 12);

      sprintf (buf, "%lu", mask);

      gtk_clist_set_pixtext (GTK_CLIST (serverinfo_clist), row, 0, buf, 4, 
                     ((flags & mask) != 0)? gplus_pix.pix : rminus_pix.pix,
                     ((flags & mask) != 0)? gplus_pix.mask : rminus_pix.mask);
    }

    flags &= ~mask;
    mask <<= 1;
  }
}


static void sort_serverinfo (char **info) {
  char **ptr;
  int n = 0;

  if (!info)
    return;

  for (ptr = info; ptr[0] || ptr[1]; ptr += 2)
    n++;

  qsort (info, n, sizeof (char *) * 2, qsort_serverinfo);
}


void populate_serverinfo_clist (struct server *s, int refresh) {
  char *text[2];
  char **info = s->info;

  if (info == NULL) {
    gtk_clist_clear (GTK_CLIST (serverinfo_clist));
    return;
  }

  gtk_clist_freeze (GTK_CLIST (serverinfo_clist));
  if (refresh) {
    while (GTK_CLIST (serverinfo_clist)->rows)
      gtk_clist_remove (GTK_CLIST (serverinfo_clist), 0);
  }
  else {
    gtk_clist_clear (GTK_CLIST (serverinfo_clist));
  }

  sort_serverinfo (info);

  while (info[0]) {
    text[0] = (info[0])? info[0] : NULL;
    text[1] = (info[1])? info[1] : NULL;

    gtk_clist_append (GTK_CLIST (serverinfo_clist), text);

    switch (s->type) {

    case QW_SERVER:
      if (info[0] && !strcmp (info[0], "teamplay")) {
	if (s->game && (!g_strcasecmp (s->game, "ctf") ||
                                       !g_strcasecmp (s->game, "starwars"))) {
	  show_extended_flags (info[1], qw_teamplay_ctf, QW_TEAMPLAY_CTF, 
                                                                       FALSE);
	}
	else if (s->game && !g_strcasecmp (s->game, "fortress")) {
	  show_extended_flags (info[1], qw_teamplay_fortress, QW_TEAMPLAY_TF, 
                                                                       FALSE);
	}
      }
      break;

    case Q2_SERVER:
      if (info[0] && !strcmp (info[0], "dmflags")) {
	if (s->game && !g_strcasecmp (s->game, "gxmod")) {
	  show_extended_flags (info[1], q2_gxmod_dmflags, Q2_GXMOD_DMFLAGS, 
                                                                       FALSE);
	}
	else if (s->game && !g_strcasecmp (s->game, "xatrix")) {
	  show_extended_flags (info[1], q2_xatrix_dmflags, Q2_XATRIX_DMFLAGS, 
                                                                       FALSE);
	}
	else if (s->game && !g_strcasecmp (s->game, "rogue")) {
	  show_extended_flags (info[1], q2_rogue_dmflags, Q2_ROGUE_DMFLAGS, 
                                                                       FALSE);
	}
	else {
	  show_extended_flags (info[1], q2_dmflags, Q2_DMFLAGS, FALSE);
	}
      }
      else if (info[0] && !strcmp (info[0], "ctfflags")) {
	if (s->game && !g_strcasecmp (s->game, "lmctf")) {
	  show_extended_flags (info[1], q2_lmctf_ctfflags, Q2_LMCTF_CTFFLAGS, 
                                                                       FALSE);
	}
      }
      else if (info[0] && !strcmp (info[0], "runes")) {
	if (s->game && !g_strcasecmp (s->game, "lmctf"))
	  show_extended_flags (info[1], q2_lmctf_runes, Q2_LMCTF_RUNES, TRUE);
      }
      else if (info[0] && !strcmp (info[0], "expflags")) {
	if (s->game && !g_strcasecmp (s->game, "expert")) {
	  show_extended_flags (info[1], q2_expert_expflags,
                                                   Q2_EXPERT_EXPFLAGS, FALSE);
	}
      }
      else if (info[0] && !strcmp (info[0], "pbflags")) {
	if (s->game && !g_strcasecmp (s->game, "pball")) {
	  show_extended_flags (info[1], q2_pball_pbflags, Q2_PBALL_PBFLAGS, 
                                                                       FALSE);
	}
      }
      else if (info[0] && !strcmp (info[0], "weapflags")) {
	if (s->game && (!g_strcasecmp (s->game, "gxmod") || 
                                             !g_strcasecmp (s->game, "gx"))) {
	  show_extended_flags (info[1], q2_gxmod_weapflags, 
                                                    Q2_GXMOD_WEAPFLAGS, TRUE);
	}
      }
      else if (info[0] && !strcmp (info[0], "powerupflags")) {
	if (s->game && !g_strcasecmp (s->game, "gxmod")) {
	  show_extended_flags (info[1], q2_gxmod_powerupflags, 
                                                 Q2_GXMOD_POWERUPFLAGS, TRUE);
	}
      }
      else if (info[0] && !strcmp (info[0], "wfflags")) {
	if (s->game && !g_strcasecmp (s->game, "wf"))
	  show_extended_flags (info[1], q2_wf_wfflags, Q2_WF_WFFLAGS, FALSE);
      }
      else if (info[0] && !strcmp (info[0], "teamflags")) {
	if (s->game && !g_strcasecmp (s->game, "bots")) {
	  show_extended_flags (info[1], q2_bots_teamflags, 
                                                    Q2_BOTS_TEAMFLAGS, FALSE);
	}
      }
      break;

    default:
      break;
    }

    info += 2;
  }

  gtk_clist_thaw (GTK_CLIST (serverinfo_clist));
}
